/* (c) Victor Yodaiken, Finite State Research LLC. 2016
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Example code for using one-way queues
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>




// start with default integer queues
#include "q.h"

void twothreads(char *testname, void *(*producer) (void *),
		void *(*consumer) (void *), owq_t * q);


// thread function types
void *iproducer(void *p);
void *iconsumer(void *p);
void *pproducer(void *p);
void *pconsumer(void *p);



// now create some double floating point queue routines
#define OWQ_ENQ enq_f
#define OWQ_DEQ deq_f
#define OWQ_INIT owq_init_f
#define OWQ_ELEMENT_T  double
#include "q.h"

//now create some queue routines for queues of an arbitrary structure
// there is an overhead here since the whole structure gets copied each
// enq or deq
typedef struct {
    int count;
    double dcount;
} packet_t;
#define OWQ_ENQ enq_p
#define OWQ_DEQ deq_p
#define OWQ_INIT owq_init_p
#define OWQ_ELEMENT_T  packet_t
#include "q.h"

 // declare the queue control structures
// these do not care about element type - the names are just documentation
owq_t iq1;
owq_t iq2;			// we'll put a longer queue here
owq_t fq;			// for double queue
owq_t pq;			// for packet queue

// each queue needs an array of the right element types - could be malloced
// or from shared memory or ...
#define SHORTIQ  1024
#define LONGIQ  (1024*1024)
#define FQ  (256)
#define PQ  (4096 +3)		//powers of 2 most efficient, but not essential

// the data for each queue is stored in an array of element type
int shorti[SHORTIQ];
int longi[LONGIQ];
double floats[FQ];
packet_t packets[PQ];


// the reason there is a special init for each queue element type is 
// to make type check work

void initializeqs()
{

    owq_init_i(&iq1, shorti, SHORTIQ);
    owq_init_i(&iq2, longi, LONGIQ);
    owq_init_f(&fq, floats, FQ);
    owq_init_p(&pq, packets, PQ);

}

int main(int argc, char **argv)
{
    int repeat_count = 1;
    if (argc > 1) {
	if (atoi(argv[1]) <= 0) {
	    fprintf(stderr, "Bad repetition count\n");
	    exit(1);
	} else
	    repeat_count = atoi(argv[1]);
    }

    initializeqs();
    while (repeat_count > 0) {
	fprintf(stdout, "Test set %d\n", repeat_count--);
	twothreads("first integer", iproducer, iconsumer, &iq1);
	twothreads("second integer", iproducer, iconsumer, &iq2);
	twothreads("packet", pproducer, pconsumer, &pq);
    }


}



volatile int poison = 0;
#define REPETITIONS (1024*1024*1020)
#define MAXSLEEP 100000

void *iproducer(void *p)
{
    int n = 0;
    int sleeps = 0;
    owq_t *q = (owq_t *) p;

    fprintf(stdout, "producer will count  %d\n", REPETITIONS);

    do {
	if (enq_i(q, n) == 0) {
	    n++;
	    sleeps = 0;
	} else {
	    if (sleeps++ > 10000) {
		usleep(1);
	    }
	}
    } while (n < REPETITIONS && sleeps < MAXSLEEP);

    if (sleeps >= MAXSLEEP) {
	fprintf(stderr, "Producer oversleeps\n");
	exit(0);
    }
    fprintf(stdout, "Producer exits after %d enqs\n", n);
    poison = 1;
    return 0;
}

void *iconsumer(void *p)
{
    int n = 0;
    int j;
    int sleeps = 0;
    owq_t *q = (owq_t *) p;

    do {
	if (deq_i(q, &j) == 0) {
	    if (j != n) {
		fprintf(stderr, "Consumer sequence error %d != %d\n",
			n, j);
		exit(0);
	    }
	    n++;
	    sleeps = 0;
	} else {
	    if (sleeps > 1000)
		usleep(1);
	    sleeps++;
	    if (poison)
		poison++;
	}
    }
    while (sleeps < MAXSLEEP && (poison < 5));

    if (sleeps >= MAXSLEEP) {
	fprintf(stderr, "Consumer oversleeps\n");
    } else {
	fprintf(stdout, "Consumer exits after %d deqs\
           Producer was %s\n", n, (poison ? "done" : "not done"));
    }
    return 0;
}


// this is lazy copy and paste programming, God forgive me.
void *pproducer(void *p)
{
    packet_t pk = { 0, };
    int sleeps = 0;
    owq_t *q = (owq_t *) p;
    fprintf(stdout, "Packet producer will count %d\n", REPETITIONS);


    do {
	if (enq_p(q, pk) == 0) {
	    pk.count++;
	    pk.dcount += 0.1;
	    sleeps = 0;
	} else {
	    if (sleeps++ > 10000) {
		usleep(1);
	    }
	}
    } while (pk.count < REPETITIONS && sleeps < MAXSLEEP);

    if (sleeps >= MAXSLEEP) {
	fprintf(stderr, "Packet Producer oversleeps\n");
	exit(0);
    }
    fprintf(stdout, "Packet Producer exits after %d enqs\n", pk.count);
    poison = 1;
    return 0;
}

void *pconsumer(void *p)
{
    packet_t pk = { 0, };
    packet_t pk1 = { 0, };
    int sleeps = 0;
    owq_t *q = (owq_t *) p;

    do {
	if (deq_p(q, &pk) == 0) {
	    if ((pk.count != pk1.count) || (pk.dcount != pk1.dcount)) {
		fprintf(stderr,
			"Consumer sequence error %d != %d %e %e\n",
			pk.count, pk1.count, pk.dcount, pk1.dcount);
		exit(0);
	    }
	    pk1.count++;
	    pk1.dcount += 0.1;
	    sleeps = 0;
	} else {
	    if (sleeps > 1000)
		usleep(1);
	    sleeps++;
	    if (poison)
		poison++;
	}
    }
    while (sleeps < MAXSLEEP && poison < 3);

    if (sleeps >= MAXSLEEP) {
	fprintf(stderr, "Packet Consumer oversleeps\n");
    } else {
	fprintf(stdout, "Consumer exits after %d deqs\
           Producer was %s\n", pk1.count, (poison ? "done" : "not done"));
    }

    return 0;
}

#include <time.h>
unsigned long millisec(void)
{
    struct timespec t;
    if (clock_gettime(CLOCK_REALTIME, &t)) {
	fprintf(stdout, "Can't read time\n");
    }

    return t.tv_sec * 1000 + ((unsigned long) t.tv_nsec) / (1000 * 1000);
}

void twothreads(char *testname, void *(*producer) (void *),
		void *(*consumer) (void *), owq_t * q)
{
    int r1, r2;
    pthread_t thread1, thread2;
    unsigned long elapsed;

    poison = 0;	// to make sure tests start right. Lame use of global.
    fprintf(stdout, "Start test %s\n", testname);
    r1 = pthread_create(&thread1, NULL, producer, (void *) q);
    r2 = pthread_create(&thread2, NULL, consumer, (void *) q);

    elapsed = millisec();

    if (r1 || r2) {
	fprintf(stdout, "Test %s thread create returns: %d %d\n",
		testname, r1, r2);
	exit(0);
    }

    pthread_join(thread1, NULL);
    pthread_join(thread2, NULL);

    if (r1 || r2) {
	fprintf(stdout, "Thread 1 returns: %d Thread 2 returns %d\n",
		r1, r2);
    }
    fprintf(stdout, "Test %s took %ld milliseconds\n",
	    testname, millisec() - elapsed);
}
